class Base64:
    BASELENGTH = 128
    LOOKUPLENGTH = 64
    TWENTYFOURBITGROUP = 24
    EIGHTBIT = 8
    SIXTEENBIT = 16
    FOURBYTE = 4
    SIGN = -128
    PAD = '='

    def __init__(self):
        self.base64Alphabet = [-1] * self.BASELENGTH
        self.lookUpBase64Alphabet = [''] * self.LOOKUPLENGTH
        self._init_alphabets()

    def _init_alphabets(self):
        for i in range(65, 91):  # A-Z
            self.base64Alphabet[i] = i - 65
        for i in range(97, 123):  # a-z
            self.base64Alphabet[i] = i - 97 + 26
        for i in range(48, 58):  # 0-9
            self.base64Alphabet[i] = i - 48 + 52
        self.base64Alphabet[43] = 62  # '+'
        self.base64Alphabet[47] = 63  # '/'

        for i in range(0, 26):
            self.lookUpBase64Alphabet[i] = chr(65 + i)
        for i in range(26, 52):
            self.lookUpBase64Alphabet[i] = chr(97 + i - 26)
        for i in range(52, 62):
            self.lookUpBase64Alphabet[i] = chr(48 + i - 52)
        self.lookUpBase64Alphabet[62] = '+'
        self.lookUpBase64Alphabet[63] = '/'

    @staticmethod
    def isWhiteSpace(char):
        return char in (' ', '\n', '\r', '\t')

    @staticmethod
    def isPad(char):
        return char == '='

    def isData(self, char):
        return ord(char) < self.BASELENGTH and self.base64Alphabet[ord(char)] != -1

    def encode(self, binaryData: bytes) -> str:
        if binaryData is None:
            return None

        lengthDataBits = len(binaryData) * self.EIGHTBIT
        fewerThan24bits = len(binaryData) % 3
        numberTriplets = len(binaryData) // 3
        encodedData = []

        dataIndex = 0
        for _ in range(numberTriplets):
            byte1 = binaryData[dataIndex]
            byte2 = binaryData[dataIndex + 1]
            byte3 = binaryData[dataIndex + 2]

            dataIndex += 3

            val1 = byte1 >> 2
            val2 = ((byte1 & 0x3) << 4) | (byte2 >> 4)
            val3 = ((byte2 & 0xf) << 2) | (byte3 >> 6)
            val4 = byte3 & 0x3f

            encodedData.extend([
                self.lookUpBase64Alphabet[val1],
                self.lookUpBase64Alphabet[val2],
                self.lookUpBase64Alphabet[val3],
                self.lookUpBase64Alphabet[val4],
            ])

        if fewerThan24bits == 1:
            byte1 = binaryData[dataIndex]
            val1 = byte1 >> 2
            val2 = (byte1 & 0x3) << 4

            encodedData.extend([
                self.lookUpBase64Alphabet[val1],
                self.lookUpBase64Alphabet[val2],
                self.PAD,
                self.PAD
            ])
        elif fewerThan24bits == 2:
            byte1 = binaryData[dataIndex]
            byte2 = binaryData[dataIndex + 1]

            val1 = byte1 >> 2
            val2 = ((byte1 & 0x3) << 4) | (byte2 >> 4)
            val3 = (byte2 & 0xf) << 2

            encodedData.extend([
                self.lookUpBase64Alphabet[val1],
                self.lookUpBase64Alphabet[val2],
                self.lookUpBase64Alphabet[val3],
                self.PAD
            ])

        return ''.join(encodedData)

    def removeWhiteSpace(self, data: str) -> str:
        return ''.join([c for c in data if not self.isWhiteSpace(c)])

    def decode(self, encoded: str) -> bytes:
        if encoded is None:
            return None

        encoded = self.removeWhiteSpace(encoded)
        if len(encoded) % 4 != 0:
            return None

        numberQuadruple = len(encoded) // 4
        decodedBytes = bytearray()

        i = 0
        for _ in range(numberQuadruple):
            c1 = encoded[i]
            c2 = encoded[i + 1]
            c3 = encoded[i + 2]
            c4 = encoded[i + 3]
            i += 4

            if not (self.isData(c1) and self.isData(c2)):
                return None

            b1 = self.base64Alphabet[ord(c1)]
            b2 = self.base64Alphabet[ord(c2)]

            b3 = self.base64Alphabet[ord(c3)] if c3 != self.PAD else 0
            b4 = self.base64Alphabet[ord(c4)] if c4 != self.PAD else 0

            byte1 = (b1 << 2) | (b2 >> 4)
            decodedBytes.append(byte1)

            if c3 != self.PAD:
                byte2 = ((b2 & 0xf) << 4) | (b3 >> 2)
                decodedBytes.append(byte2)

                if c4 != self.PAD:
                    byte3 = ((b3 & 0x3) << 6) | b4
                    decodedBytes.append(byte3)

        return bytes(decodedBytes)

class Base64Utils:
    @staticmethod
    def encode(data: bytes) -> str:
        return Base64().encode(data).decode("utf-8")

    @staticmethod
    def decode(data: str) -> bytes:
        return Base64().decode(data.encode("utf-8"))

